%{
// Author:		Dario Accornero <adario@cs.bu.edu>
// BeDev ID:	10183
// Project:		BeScript
// Module:
// Lexical analyzer.

/*
	Copyright 2000, Dario Accornero.   All Rights Reserved.
	This file may be used under the terms of the Modified Be Sample Code License
	(the "Software License" document accompanying this file).
*/

// System headers.

#ifndef __CSTDIO__
#include <cstdio>
#endif
#ifndef __CSTDLIB__
#include <cstdlib>
#endif

// Local headers.

#include "errormsg.h"
#include "yystype.h"
#include "y.tab.cc.h"

// MACROS:
// Avoid using yyunput(), since it is not needed by our grammar.

#define YY_NO_UNPUT 1

// We declare our own version of yywrap().

#define YY_SKIP_YYWRAP 1

// We don't want empty lines to be echoed.

#define ECHO { if (yyleng && (*yytext != '\n' && *yytext != '\r')) (void) fwrite( yytext, yyleng, 1, yyout ); }

// LOCAL DATA:
// Reading position in current line.

static int    charPos = 1;

// Temporary string to read a string constant from input.

static string strLiteral = "";

// Keep track of comments nesting.

static int    commentLevel = 0;

// LOCAL FUNCTIONS:
// Reset reading position.
// This is not declared static since it's used by other modules.

int yywrap(void);

// Advance reading position.
// Note that advancing over TAB characters is terminal-dependent when it comes
// to displaying error messages.

static void adjust(void);

%}

  /* Macros. */
  /* We only need few of them, since fortunately the language structure */
  /* is rather straightforward. Currently, the floating point macro */
  /* is not particularly smart but it should suffice for most purposes. */
  /* Identifiers are alphanumeric strings that do not begin with a digit. */

control				[0-9][0-9][0-9]
integer				[0-9]+
floatingp			[0-9]*\.[0-9]+
identifier			[a-zA-Z_][a-zA-Z_0-9]*

  /* Start states. INITIAL is defined by default. */

%start COMMENT STRINGC POUNDLINE

%%

  /* The analysis shall begin from the first rule of the INITIAL state. */

  /* Begin pound line, as in initial shell invokation from a script. */
  /* This also means that any line starting with a pound sign will be */
  /* considered like a comment, e.g. ignored. */
  
<INITIAL>#			{ adjust(); BEGIN POUNDLINE; }
<POUNDLINE>[\n\r]+	{ adjust(); BEGIN INITIAL; }
<POUNDLINE>.		{ adjust(); continue; }

  /* Begin comment. */

<INITIAL>"/*"		{ adjust(); ++commentLevel; BEGIN COMMENT; }

  /* Comment handling. */

<COMMENT>"*/"		{	adjust();
						if (--commentLevel < 0)
						{
							EM_error(EM_tokPos, "badly nested comment");
							continue;
						}
						else
							BEGIN INITIAL;	}
<COMMENT>[\n\r]		{ adjust(); EM_newline(); continue; }
<COMMENT>.			{ adjust(); continue; }

  /* Whitespace. */

<INITIAL>" "		{ adjust(); continue; }
<INITIAL>\t			{ adjust(); continue; }
<INITIAL>\n			{ adjust(); EM_newline(); continue; }
<INITIAL>\r			{ adjust(); EM_newline(); continue; }

  /* Punctuation symbols and operators. */
  /* We include in quotes most of them in order to specify them directly. */
  /* Currently we use only a few operators. */

<INITIAL>"."		{ adjust(); return DOT; }
<INITIAL>","		{ adjust(); return COMMA; }
<INITIAL>":"		{ adjust(); return COLON; }
<INITIAL>";"		{ adjust(); return SEMICOLON; }
<INITIAL>"("		{ adjust(); return LPAREN; }
<INITIAL>")"		{ adjust(); return RPAREN; } 
<INITIAL>"["		{ adjust(); return LBRACK; }
<INITIAL>"]"		{ adjust(); return RBRACK; }
<INITIAL>"{"		{ adjust(); return LBRACE; }
<INITIAL>"}"		{ adjust(); return RBRACE; }
<INITIAL>"+"		{ adjust(); return PLUS; }
<INITIAL>"-"		{ adjust(); return MINUS; }
<INITIAL>"*"		{ adjust(); return TIMES; }
<INITIAL>"/"		{ adjust(); return DIVIDE; }
<INITIAL>"=="		{ adjust(); return EQ; }
<INITIAL>"!="		{ adjust(); return NEQ; }
<INITIAL>"<="		{ adjust(); return LE; }
<INITIAL>"<"		{ adjust(); return LT; }
<INITIAL>">="		{ adjust(); return GE; }
<INITIAL>">"		{ adjust(); return GT; }
<INITIAL>"="		{ adjust(); return ASSIGNMENT; }
<INITIAL>and		{ adjust(); return AND; }
<INITIAL>or			{ adjust(); return OR; }

  /* Command constants. From 4 to 1 characters. */
  
<INITIAL>\'....\'	{	unsigned char* text = (unsigned char*)yytext;
						adjust();
						yylval.wval =	(((uint32)text[1]) << 24) |
										(((uint32)text[2]) << 16) |
										(((uint32)text[3]) << 8)  |
										(((uint32)text[4]));
						return VERBID; }
<INITIAL>\'...\'	{	unsigned char* text = (unsigned char*)yytext;
						adjust();
						yylval.wval = 0;
						yylval.wval =	(((uint32)text[1]) << 16) |
										(((uint32)text[2]) << 8)  |
										(((uint32)text[3]));
						return VERBID; }
<INITIAL>\'..\'		{	unsigned char* text = (unsigned char*)yytext;
						adjust();
						yylval.wval = 0;
						yylval.wval =	(((uint32)text[1]) << 8)  |
										(((uint32)text[2]));
						return VERBID; }
<INITIAL>\'.\'		{	unsigned char* text = (unsigned char*)yytext;
						adjust();
						yylval.wval = (uint32)text[1];
						return VERBID; }

  /* A single quote. */
  
<INITIAL>\'			{ adjust(); return SINGLEQUOTE; }

  /* Reserved keywords. */
  /* These are needed to represent the actual statements of the language. */
  /* We list abbreviations too. */

<INITIAL>getsuites	{ adjust(); return GETSUITES; }
<INITIAL>get		{ adjust(); return GET; }
<INITIAL>set		{ adjust(); return SET; }
<INITIAL>create		{ adjust(); return CREATE; }
<INITIAL>execute	{ adjust(); return EXECUTE; }
<INITIAL>exec		{ adjust(); return EXECUTE; }
<INITIAL>delete		{ adjust(); return DELETE; }
<INITIAL>del		{ adjust(); return DELETE; }
<INITIAL>count		{ adjust(); return COUNT; }
<INITIAL>tell		{ adjust(); return TELL; }
<INITIAL>endtell	{ adjust(); return ENDTELL; }
<INITIAL>endt		{ adjust(); return ENDTELL; }
<INITIAL>quit		{ adjust(); return QUIT; }
<INITIAL>about		{ adjust(); return ABOUT; }
<INITIAL>save		{ adjust(); return SAVE; }
<INITIAL>load		{ adjust(); return LOAD; }

  /* Raw specifiers interaction. */
  /* The obnoxious "of" is contained in the next section. */
  /* The "property", "value", "message", and "result" keywords are currently unused, */
  /* and they must remain lowercase only. */
  /* Truth values are available in UPPERCASE, MixedCase and lowercase. */

<INITIAL>the		{ adjust(); return THE; }
<INITIAL>property	{ adjust(); return PROPERTY; }
<INITIAL>with		{ adjust(); return WITH; }
<INITIAL>result		{ adjust(); return RESULT; }
<INITIAL>value		{ adjust(); return VALUE; }
<INITIAL>message	{ adjust(); return MESSAGE; }
<INITIAL>BRect		{ adjust(); return BRECT; }
<INITIAL>BPoint		{ adjust(); return BPOINT; }
<INITIAL>rgb_color	{ adjust(); return RGBCOLOR; }
<INITIAL>bool		{ adjust(); return BOOL; }
<INITIAL>string		{ adjust(); return STRING; }
<INITIAL>int8		{ adjust(); return INT8; }
<INITIAL>int16		{ adjust(); return INT16; }
<INITIAL>int32		{ adjust(); return INT32; }
<INITIAL>float		{ adjust(); return FLOAT; }
<INITIAL>double		{ adjust(); return DOUBLE; }
<INITIAL>file		{ adjust(); return BFILE; }
<INITIAL>FALSE		{ adjust(); yylval.bval = false; return BFALSE; }
<INITIAL>False		{ adjust(); yylval.bval = false; return BFALSE; }
<INITIAL>false		{ adjust(); yylval.bval = false; return BFALSE; }
<INITIAL>TRUE		{ adjust(); yylval.bval = true; return BTRUE; }
<INITIAL>True		{ adjust(); yylval.bval = true; return BTRUE; }
<INITIAL>true		{ adjust(); yylval.bval = true; return BTRUE; }

  /* Additional commands and keywords. */
  
<INITIAL>launch		{ adjust(); return LAUNCH; }
<INITIAL>click		{ adjust(); return CLICK; }
<INITIAL>button		{ adjust(); return BUTTON; }
<INITIAL>type		{ adjust(); return TYPE; }
<INITIAL>into		{ adjust(); return INTO; }
<INITIAL>field		{ adjust(); return FIELD; }
<INITIAL>choose		{ adjust(); return CHOOSE; }
<INITIAL>pause		{ adjust(); return PAUSE; }
<INITIAL>print		{ adjust(); return PRINT; }
<INITIAL>assign		{ adjust(); return ASSIGN; }

  /* Control keywords. */
  /* Some of them may remain unused and will be removed if necessary. */

<INITIAL>if			{ adjust(); return IF; }
<INITIAL>then		{ adjust(); return THEN; }
<INITIAL>else		{ adjust(); return ELSE; }
<INITIAL>while		{ adjust(); return WHILE; }
<INITIAL>for		{ adjust(); return FOR; }
<INITIAL>to			{ adjust(); return TO; }
<INITIAL>do			{ adjust(); return DO; }
<INITIAL>let		{ adjust(); return LET; }
<INITIAL>in			{ adjust(); return IN; }
<INITIAL>end		{ adjust(); return END; }
<INITIAL>of			{ adjust(); return OF; }
<INITIAL>break		{ adjust(); return BREAK; }
<INITIAL>nil		{ adjust(); return NIL; }

  /* Identifiers. */
  /* Currently we use them only for application targets. */

<INITIAL>{identifier}	{ 	adjust();
                          	yylval.sval = string(yytext);
                          	return ID; }

  /* Numeric literals. */

<INITIAL>{integer}		{ 	adjust();
							yylval.ival = atoi(yytext);
							return INT; }
<INITIAL>{floatingp}	{ 	adjust();
							yylval.fval = atof(yytext);
							return FP; }

  /* String literals. */
  /* They will be manipulated to convert escape sequences and control characters. */
  /* Literals must be included in quotes -- apostrophe/backtick not supported. */

<INITIAL>\"				{	adjust();
							yylval.sval = string("");
							strLiteral = string("");
					  		BEGIN STRINGC; }
<STRINGC>\\[ \n\r\t]+\\	{	adjust(); continue; }
<STRINGC>\\n			{	adjust(); strLiteral += "\n"; continue; }
<STRINGC>\\r			{	adjust(); strLiteral += "\r"; continue; }
<STRINGC>\\t			{	adjust(); strLiteral += "\t"; continue; }
<STRINGC>\\\"			{	adjust(); strLiteral += "\""; continue; }
<STRINGC>\\{control}	{	char cntrl[2] = {0, 0};
			  				cntrl[0] = (char)atoi(&yytext[1]);
			  				strLiteral += cntrl;
			  				adjust();
			  				continue; }
<STRINGC>\\^[A-Z]		{	char cntrl[2] = {0, 0};
							cntrl[0] = (yytext[2] - 'A') + 1;
							strLiteral += cntrl;
			  				adjust();
			  				continue; }
<STRINGC>\\				{	adjust(); strLiteral += "\\"; continue; }
<STRINGC>\"				{	adjust();
			  				yylval.sval = strLiteral;
			  				BEGIN INITIAL;
			  				return STRINGID; }
<STRINGC>.				{	adjust();
			  				strLiteral += yytext;
			  				continue; }

  /* Any other input. */
  /* If we reach this point, we have an error. */
  /* We cannot do error recovery from the lexer -- if we ever get to that point, */
  /* we'll do it in the grammar specification. */

<INITIAL>.				{	adjust();
							ECHO;
							EM_error(EM_tokPos, "illegal token"); }

%%

// LOCAL FUNCTIONS:
// Reset reading position.

int yywrap(void)
{
	charPos = 1;
	return 1;
}

// Advance reading position.

static void adjust(void)
{
	EM_tokPos = charPos;
	charPos += yyleng;
}
